﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using connector = Microsoft.Office.Server.Search.Connector;
using Microsoft.Office.Server.Search.Connector.BDC;
using Microsoft.BusinessData.MetadataModel;
using System.IO;
using Microsoft.SharePoint.Utilities;
using Microsoft.SharePoint;
using XmlFileConnector.Utils;
using Microsoft.Office.Server.Search.Connector;
using Microsoft.BusinessData.MetadataModel.Collections;
using Microsoft.Office.Server.Search.Administration;
using Microsoft.SharePoint.Administration;

namespace XmlFileConnector
{
    /// <summary>
    /// Defines methods that map the URL coming in from search to the Object in Metadata store
    /// 
    /// <remarks>
    /// When creating a new Content source using this connector, the start url of the content source has to be formatted as 
    /// shown in the following sample:
    ///     - xmldoc://demo2010a/c$/content/#x=:doc:url;;urielm=url;;titleelm=title#
    /// where:
    ///     - "xmldoc" is the name of name of the protocol of this connector as registered in the Registry when installing this connector
    ///     - "//demo2010a/c$/content/" is the full path to directory holding our xml files that are to be crawled
    ///     - "#x=:doc:url;;urielm=url;;titleelm=title#" is a special way to encode config parameters used by the connector
    ///         - "x=:doc:url" is the path of element in the xml document that specifies the id of a single item
    ///         - "urielm=url" is the name of the element in the xml document that will be set as the uri of the crawled item
    ///         - "titleelm=title" is the name of the element in the xml document that will be set as the title of the crawled item
    /// </remarks>
    /// </summary>
    class XmlDocumentLobUri : LobUri
    {
        #region Constants

        /// <summary>
        /// xmldoc is the scheme of the URIs
        /// </summary>
        public const string XMLDocScheme = "xmldoc";
        
        /// <summary>
        /// Name of LobSystem as defined in model file
        /// </summary>
        private const string LobSystemName = "XmlDocumentSystem";

        private const string XmlDocumentConnectorNameSpace = "XmlDocumentConnector";
        private const string XmlDocumentEntityName = "XmlDocument";
        private const string FolderEntityName = "Folder";

        #endregion

        public XmlDocumentLobUri()
            : base(XMLDocScheme) 
        {
            // 
            this.lobSystem = this.Catalog.GetLobSystem(LobSystemName); 
            
            // We have only one LobSystemInstance defined in our model file
            this.lobSystemInstance = this.lobSystem.GetLobSystemInstances()[0].Value;
        }
                
        public override void Initialize(IConnectionContext context)
        {
            Uri startAddress = context.Path;

            string originalUrlString = SPHttpUtility.UrlPathDecode(startAddress.OriginalString, false);

            /* Parse the url string to extract the components
             * the url string may look something like this:
             * 
             *      xmldoc://demo2010a/c$/content/#x=:doc:url;;urielm=url;;titleelm=title#/documentid
             * 
             * This will result in the following components:
             *  - folderpath: //demo2010a/c$/content/
             *  - configParams: x=:doc:url;;urielm=url;;titleelm=title
             *  - docid: /documentid
             */

            LoggingService.LogVerbose("Original Start Address: {0}", originalUrlString);

            string[] parts = originalUrlString.Split(new string[]{ "#" }, StringSplitOptions.RemoveEmptyEntries);
            
            string folderpath = "";
            string configParams = "";
            string docid = "";

            if (parts.Length >= 2)
            {
                folderpath = parts[0].Replace("xmldoc:", "").Replace("/", "\\");
                configParams = parts[1];
            }

            if (parts.Length == 3)
            {
                if(!String.IsNullOrEmpty(parts[2].Trim()))
                    docid = parts[2].Replace("/", "");
            }
                        
            if (!String.IsNullOrEmpty(docid))
            {
                // We got a document id as part of the url sent in by the BCS framework
                // Set the entity to XmlDocument

                LoggingService.LogVerbose("Extracted docid: {0} - returning {1} as entity.", docid, XmlDocumentEntityName);

                this.entity = this.Catalog.GetEntity(XmlDocumentConnectorNameSpace, XmlDocumentEntityName);
                this.identity = new Microsoft.BusinessData.Runtime.Identity(docid);
            }
            else if (!String.IsNullOrEmpty(folderpath) && Directory.Exists(folderpath))
            {
                // Still traversing the configured folder path
                // Set the entity to Folder

                LoggingService.LogVerbose("Extracted folder: {0} - returning {1} as entity.", folderpath, FolderEntityName);

                this.entity = this.Catalog.GetEntity(XmlDocumentConnectorNameSpace, FolderEntityName);
                this.identity = new Microsoft.BusinessData.Runtime.Identity(folderpath);
            }
        }

        #region Properties

        private Uri sourceUri;
        public override Uri SourceUri
        {
            get
            {
                return this.sourceUri;
            }
            set
            {
                this.sourceUri = value;
            }
        }

        private IEntity entity;
        public override Microsoft.BusinessData.MetadataModel.IEntity Entity
        {
            get { return this.entity; }
        }

        private Microsoft.BusinessData.Runtime.Identity identity;
        public override Microsoft.BusinessData.Runtime.Identity Identity
        {
            get { return this.identity; }
        }

        private ILobSystem lobSystem;
        public override Microsoft.BusinessData.MetadataModel.ILobSystem LobSystem
        {
            get { return this.lobSystem; }
        }

        private ILobSystemInstance lobSystemInstance;
        public override Microsoft.BusinessData.MetadataModel.ILobSystemInstance LobSystemInstance
        {
            get { return this.lobSystemInstance; }
        }

        public override Guid PartitionId
        {
            get { return Guid.Empty; }
        }

        #endregion
    }
}
