﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.IO;
using System.Security.AccessControl;
using Microsoft.BusinessData.Runtime;
using XmlFileConnector.Utils;
using System.Text;
using System.Linq;
using System.Xml.Linq;
using Microsoft.SharePoint;
using Microsoft.SharePoint.Administration;
using Microsoft.Office.Server.Search.Administration;
using System.Globalization;

namespace XmlFileConnector
{
    /// <summary>
    /// Implements the methods that return data from the XMl files
    /// </summary>
    public class XmlDocumentProxy : IDisposable
    {
        /// <summary>
        /// Id of this crawl
        /// </summary>
        internal string CrawlId { get; set; }
        internal string XmlInputFolderPath { get; set; }
        internal string DocumentAndIDPath { get; set; }
        internal string UriElementName { get; set; }
        internal string TitleElementName { get; set; }

        private void Initialize()
        {
            // initialize the XmlFileLoader instance for this run
            XmlFileLoader xmlFileLoader = XmlFileLoader.GetInstance(this.CrawlId);
            
            // make sure we clear the caches so that we start clean for this run
            xmlFileLoader.ClearCaches();
        }

        /// <summary>
        /// Returns the start address as the root folder
        /// </summary>
        /// <returns></returns>
        [Browsable(true)]
        public FolderBatch GetRootFolders()
        {
            LoggingService.LogVerbose("Running GetRootFolders() for Input Folder Path: {0}", this.XmlInputFolderPath);

            List<Folder> folders = new List<Folder>();

            Folder folder = new Folder();
            folder.Path = this.XmlInputFolderPath;
            DirectoryInfo di = new DirectoryInfo(this.XmlInputFolderPath);
            folder.LastModifiedTime = di.LastWriteTimeUtc;

            folders.Add(folder);

            FolderBatch folderBatch = new FolderBatch();
            folderBatch.Folders = folders.ToArray();
                        
            LoggingService.LogVerbose("GetRootFolders() - Returning FolderBatch with {0} folders.", folderBatch.Folders.Length);

            return folderBatch;
        }

        [Browsable(true)]
        public Folder GetSpecifiedFolder(string folderpath)
        {
            LoggingService.LogVerbose("GetSpecifiedFolder(folderpath={0})", folderpath);

            DirectoryInfo di = new DirectoryInfo(folderpath);
            if (di.Exists)
            {
                // make sure we start clean here, since this would be a new crawl
                this.Initialize();

                Folder folder = new Folder();
                folder.Path = folderpath;
                folder.LastModifiedTime = di.LastWriteTimeUtc;

                // TODO: make this read files according to date
                // and make this also read .del files
                IEnumerable<FileInfo> xmlFiles = di.GetFiles("*.xml").OrderBy(f => f.LastWriteTime);
                foreach (var xf in xmlFiles)
                {
                    XmlFileLoader.GetInstance(this.CrawlId).EnqueueXmlFile(xf.FullName);
                    LoggingService.LogInfo("Enqueued {0} for processing.", xf.FullName);
                }

                IEnumerable<FileInfo> delFiles = di.GetFiles("*.del").OrderBy(f => f.LastWriteTime);
                foreach (var df in delFiles)
                {
                    XmlFileLoader.GetInstance(this.CrawlId).AddDeletionFile(df);
                    LoggingService.LogInfo("Enqueued {0} for processing.", df.FullName);
                }
                
                return folder;
            }

            return null;
        }

        
        /// <summary>
        /// The finder method implementation that returns all objects
        /// </summary>
        /// <returns>All documents for the given start address</returns>
        [Browsable(true)]
        public DocumentBatch GetAllDocuments(string folderpath, int offset)
        {
            LoggingService.LogVerbose("GetAllDocuments(folderPath={0}, offSet={1})", folderpath, offset);

            bool hasMore = false;

            string xmlFileName = XmlFileLoader.GetInstance(this.CrawlId).DequeueXmlFile(out hasMore);

            if (xmlFileName != null)
            {
                LoggingService.LogVerbose("Getting documents from: {0}", xmlFileName);
                LoggingService.LogVerbose("Are there more documents following this batch? {0}", hasMore);

                List<Document> docs = XmlFileLoader.GetInstance(this.CrawlId).GetDocumentsFromFile(xmlFileName, this.DocumentAndIDPath);

                DocumentBatch docBatch = new DocumentBatch
                {
                    Documents = docs.ToArray(),
                    IncludesLastItemInRange = !hasMore,
                    Offset = docs.Count
                };

                LoggingService.LogVerbose("Returning document batch with {0} documents.", docBatch.Documents.Length);

                return docBatch;
            }

            // return an empty batch in case no files could be retrieved
            DocumentBatch emptyBatch = new DocumentBatch
            {
                Documents = new Document[0] { },
                IncludesLastItemInRange = true,
                Offset = 0
            };

            return emptyBatch;
        }

        /// <summary>
        /// Implementation of the SpecificFinder stereotype
        /// </summary>
        /// <param name="documentId">The id of the document to return.</param>
        /// <returns>The document object.</returns>
        [Browsable(true)]
        public Document GetSpecifiedDocument(string documentId)
        {
            Document doc = XmlFileLoader.GetInstance(this.CrawlId).GetSpecificDocument(documentId, this.UriElementName, this.TitleElementName);
            
            return doc;
        }

        /// <summary>
        /// Implementation of the streamaccessor stereotype
        /// </summary>
        /// <param name="documentId">The id of the document to return.</param>
        /// <returns>A stream to the xml data.</returns>
        [Browsable(true)]
        public Stream GetDocumentData(string documentId)
        {
            XElement elm = XmlFileLoader.GetInstance(this.CrawlId).GetSpecificDocumentAsXml(documentId);

            byte[] bytes = Encoding.UTF8.GetBytes(elm.ToString());

            MemoryStream memoryStream = new MemoryStream();
            memoryStream.Write(bytes, 0, bytes.Length);
            memoryStream.Seek(0, SeekOrigin.Begin);

            return memoryStream;
        }

        public void SetProxyServerInfo(
           string proxyServerName,
           string bypassList,
           bool bypassProxyForLocalAddress)
        {
        }

        #region IDisposable Members

        public void Dispose() { }

        #endregion
    }
}
