﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml.Linq;
using XmlFileConnector.Utils;
using System.IO;
using Microsoft.SharePoint.Utilities;
using Microsoft.Office.Server.Search.Administration;

namespace XmlFileConnector
{
    /// <summary>
    /// Helper class that handles reading from xml files and caching the content.
    /// </summary>
    internal class XmlFileLoader
    {
        private static Dictionary<string, XmlFileLoader> _instanceCache = new Dictionary<string, XmlFileLoader>();
        private static object lockObject = new object();
        
        private Dictionary<string, XElement> DocumentCache = new Dictionary<string, XElement>();

        private object fileNameCacheLock = new object();
        private Queue<string> FileNameCache = new Queue<string>();

        private List<FileInfo> DeletionFileCache = new List<FileInfo>();
        
        protected XmlFileLoader()
        { }
        
        /// <summary>
        /// Returns a cached instance of XMLFileLoader with the given id.
        /// This to handle the case when multiple Content Sources are created, 
        /// each Content Source should get its own instance of this class.
        /// </summary>
        /// <param name="id">Id of the XMLFileLoaded instance.</param>
        /// <returns>An instance of XMLFileLoader</returns>
        public static XmlFileLoader GetInstance(string id)
        { 
            if(_instanceCache.ContainsKey(id))
                return _instanceCache[id];

            lock (lockObject)
            {
                if (!_instanceCache.ContainsKey(id))
                {
                    _instanceCache.Add(id, new XmlFileLoader());
                }
                
                return _instanceCache[id];
            }
        }

        /// <summary>
        /// Clears all caches.
        /// </summary>
        public void ClearCaches()
        {
            lock (fileNameCacheLock)
            {
                this.FileNameCache.Clear();
            }

            this.DocumentCache.Clear();
        }

        public void AddDeletionFile(FileInfo delFile)
        {
            lock (fileNameCacheLock)
            {
                this.DeletionFileCache.Add(delFile);
            }
        }

        /// <summary>
        /// Add an XML file for processing.
        /// </summary>
        /// <param name="xmlFileName">Name of the XML file.</param>
        public void EnqueueXmlFile(string xmlFileName)
        {
            lock (fileNameCacheLock)
            {
                this.FileNameCache.Enqueue(xmlFileName);
            }
        }

        /// <summary>
        /// Returns an XML file name from the queue of files to process.
        /// </summary>
        /// <param name="hasMore">A boolean indicating whether there are still files available 
        /// in the queue after this call or not. This is used for batching.</param>
        /// <returns>Name of an XML file.</returns>
        public string DequeueXmlFile(out bool hasMore)
        {
            hasMore = false;

            lock (fileNameCacheLock)
            {
                if (FileNameCache.Count > 0)
                {
                    string xmlFileName = this.FileNameCache.Dequeue();
                    if (FileNameCache.Count > 0)
                    {
                        hasMore = true;
                    }

                    return xmlFileName;
                }

                return null;
            }
        }

        public List<string> GetDeletedDocumentIdsForXmlFile(string xmlFileName)
        {
            List<string> deletedIds = new List<string>();

            FileInfo xmlFile = new FileInfo(xmlFileName);

            if (xmlFile.Exists)
            {
                var deletionFiles = DeletionFileCache.Where(f => f.LastWriteTime > xmlFile.LastWriteTime);
                foreach (var delFile in deletionFiles)
                {
                    if (delFile.Exists)
                    {
                        int counter = 0;

                        using (StreamReader sr = delFile.OpenText())
                        {
                            string line = null;
                            while ((line = sr.ReadLine()) != null)
                            {
                                deletedIds.Add(line.Trim());
                            }
                        }
                    }
                }
            }

            return deletedIds;
        }

        /// <summary>
        /// Reads the contents of the XML file and returns a list containing all the documents in the file.
        /// </summary>
        /// <param name="xmlFileName">Name of the XML file to read.</param>
        /// <param name="documentAndIdPath">Configuration parameter used when identifying document and id elements.</param>
        /// <returns></returns>
        public List<Document> GetDocumentsFromFile(string xmlFileName, string documentAndIdPath)
        {
            List<Document> docs = new List<Document>();

            FileInfo fi = new FileInfo(xmlFileName);
            if (fi.Exists)
            {
                List<string> deletedDocumentIds = GetDeletedDocumentIdsForXmlFile(xmlFileName);

                DateTime fileModifiedTime = fi.LastWriteTime;

                string documentElmName = "doc"; // default document element name
                string documentIdElmName = "id"; // default id element name

                // try to parse the element names to use
                if (!String.IsNullOrEmpty(documentAndIdPath))
                {
                    string[] parts = documentAndIdPath.Split(new string[] { ":" }, StringSplitOptions.RemoveEmptyEntries);
                    if (parts.Length >= 2)
                    {
                        if(!String.IsNullOrEmpty(parts[0]))
                            documentElmName = parts[0];
                        
                        if (!String.IsNullOrEmpty(parts[1]))
                        documentIdElmName = parts[1];
                    }
                }

                XElement xElem = XElement.Load(xmlFileName);
                foreach (XElement elm in xElem.Elements(documentElmName))
                {
                    if (!elm.HasElements)
                    {
                        LoggingService.LogWarning("XML document element does not have any child elements! Skipping.");
                        continue;
                    }
                    
                    // try to extract the document id
                    string id = null;

                    if (elm.Element(documentIdElmName) != null
                        && !String.IsNullOrEmpty(elm.Element(documentIdElmName).Value))
                    {
                        id = elm.Element(documentIdElmName).Value.Trim();
                    }
                    else if (elm.Attribute(documentIdElmName) != null
                        && !String.IsNullOrEmpty(elm.Attribute(documentIdElmName).Value))
                    {
                        id = elm.Attribute(documentIdElmName).Value.Trim();
                    }
                    else
                    { 
                        // No id element nor attribute found
                        LoggingService.LogWarning("Could not find id element or attribute for XML document! Skipping whole element.");
                        continue;
                    }

                    if (deletedDocumentIds.Contains(id))
                    {
                        LoggingService.LogInfo("XML document with id {0} has is marked as being deleted by a post deletion file! Skipping this document.", id);
                        continue;
                    }

                    elm.Add(new XElement("LAST_MODIFIED_TIME", fileModifiedTime.Ticks));
                    elm.Add(new XElement("XFC_ORIGINAL_ID", id));

                    // create our own id which is used as an identifier for the document into our cache
                    id = MD5.CalculateMD5Hash(id);

                    elm.Add(new XElement("XFC_GENERATED_ID", id));
                    
                    // cache the xml document so that we do not need to read it from file again for next read
                    DocumentCache.Add(id, elm);
                    
                    Document d = new Document()
                    {
                        DocumentID = id,
                        LastModifiedTime = fileModifiedTime
                    };

                    docs.Add(d);
                }
            }

            return docs;
        }

        /// <summary>
        /// Gets a specific document from the cache.
        /// </summary>
        /// <param name="id"></param>
        /// <param name="uriElementName">Configuration parameter defining the element used as the Uri.</param>
        /// <param name="titleElementName">Configuration parameter defining the element used as the Title.</param>
        /// <returns></returns>
        public Document GetSpecificDocument(string id, string uriElementName, string titleElementName)
        {
            if (DocumentCache.ContainsKey(id))
            {
                XElement elm = DocumentCache[id];

                string uriElm = "url"; // default value
                string titleElm = "title"; // default value
                if (!String.IsNullOrEmpty(uriElementName))
                {
                    uriElm = uriElementName;
                }

                if (!String.IsNullOrEmpty(titleElementName))
                {
                    titleElm = titleElementName;
                }

                DateTime lastModifiedTime = DateTime.Now;
                long ticks = 0;
                if (long.TryParse(elm.Element("LAST_MODIFIED_TIME").Value, out ticks))
                {
                    lastModifiedTime = new DateTime(ticks);
                }

                List<DocumentProperty> documentProps = new List<DocumentProperty>();

                documentProps.Add(new DocumentProperty() 
                                 { 
                                    PropertyName = "XMLContent",
                                    PropertyValue = elm.ToString()
                                 });

                // add all xml elements in the document as "dynamic" properties
                // these element names will show up as crawled properties
                foreach (var subElm in elm.Elements())
                {
                    documentProps.Add(new DocumentProperty() 
                                        { 
                                            PropertyName = subElm.Name.LocalName, 
                                            PropertyValue = subElm.Value
                                        });
                    LoggingService.LogVerbose(String.Format("Found property: {0}/{1}", subElm.Name.LocalName, subElm.Value));
                }
                
                Document d = new Document()
                {
                    DocumentID = id,
                    LastModifiedTime = lastModifiedTime,
                    Title = (elm.Element(titleElm) != null ? elm.Element(titleElm).Value : "notitle"), // default title value
                    Url = (elm.Element(uriElm) != null ? elm.Element(uriElm).Value : String.Format("xmldoc://id/{0}", elm.Element("XFC_ORIGINAL_ID").Value)), // default url value
                    DocumentProperties = documentProps.ToArray()
                };
                
                return d;
            }
            else
            {
                LoggingService.LogError("Unexpected cache miss for document with id: {0}. Returning null.", id);

                return null;
            }
        }

        /// <summary>
        /// Gets a specified document from the chache as XML.
        /// This is used by the StreamAccessor.
        /// </summary>
        /// <param name="id"></param>
        /// <returns></returns>
        public XElement GetSpecificDocumentAsXml(string id)
        {
            if (DocumentCache.ContainsKey(id))
            {
                XElement elm = DocumentCache[id];

                // clear the doc from the cache
                DocumentCache.Remove(id);

                return elm;
            }

            return null;
        }
    }
}